#include <glib.h>

#include <utillib.h>

#include "linelog.h"
#include "debug.h"

GType           linelog_get_type       ();
static void     linelog_class_init     (LinelogClass *klass);
static void     linelog_init           (Linelog *linelog);

static gpointer parent_class = NULL;

GType linelog_get_type ()
{
    static GType linelog_type = 0;

    if (!linelog_type)
    {
        static const GTypeInfo linelog_info =
        {
            sizeof (LinelogClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) linelog_class_init,
            NULL,           /* class_finalize */
            NULL,           /* class_data */
            sizeof (Linelog),
            0,              /* n_preallocs */
            (GInstanceInitFunc) linelog_init
        };

        linelog_type = g_type_register_static (G_TYPE_OBJECT,
                                                       "Linelog",
                                                       &linelog_info,
                                                       0);
    }

    return linelog_type;
}

static void linelog_class_init (LinelogClass *klass)
{
    parent_class = g_type_class_peek_parent (klass);
}

static void linelog_init (Linelog *linelog)
{
	linelog->begin = 0;
	linelog->end = 0;
    linelog->last_seen = 0;
    linelog->last_written = 0;

    linelog->path = 0;

    linelog->disklog = 0;

    linelog->logfile = 0;
}

Linelog *linelog_new()
{
    return LINELOG (g_object_new (linelog_get_type (), NULL));
}

/* This function adds an element to the list.
 */
void
linelog_add (Linelog *linelog, GString *line, GError **error)
{
    GIOStatus status;
    GError *err = NULL;
    gsize bytes_written;

    /* The first time we add an element we initialize the end */
    if (linelog->begin == NULL)
    {
        linelog->begin = g_list_prepend (linelog->begin, line);
        linelog->end = linelog->begin;
    }
    /* the rest of the calls are just g_list_prepend () */
    else
    {
        linelog->begin = g_list_prepend (linelog->begin, line);
    }

    /* if we have a logfile, write the line to it */
    if (linelog->logfile != NULL)
    {
        status = g_io_channel_write_chars(linelog->logfile,
                                          line->str,
                                          line->len,
                                          &bytes_written,
                                          &err);

        if (err != NULL)
        {
            g_propagate_error (error, err);
            g_error_free (err);
            return;
        }

        status = g_io_channel_flush (linelog->logfile, &err);
        if (err != NULL)
        {
            g_propagate_error (error, err);
            g_error_free (err);
            return;
        }
    }
}

/* This function returns the next element to the last returned by this
   same function.
 */
GList*
linelog_get (Linelog *linelog)
{
    if (linelog->last_seen == NULL)
    {
        /* This happens when there are no elements in the list */
        if (linelog->end == NULL)
        {
            return NULL;
        }
        /* This happens the first time we return a non-NULL value,
           i.e. the last element of the list */
        else
        {
            linelog->last_seen = linelog->end;
            return linelog->last_seen;
        }
    }
    else
    {
        /* There is no new line to return since the last time this
           function was called */
        if (linelog->last_seen->prev == NULL)
        {
            return NULL;
        }
        /* Return the new line that was added since last call */
        else
        {
            linelog->last_seen = linelog->last_seen->prev;
            return linelog->last_seen;
        }
    }
}

GList*
linelog_get_end (Linelog *linelog)
{
    return linelog->end;
}

void
linelog_setfile (Linelog *linelog, gchar *path)
{
    g_return_if_fail (path != NULL);

    if (linelog->path != NULL)
    {
        g_free (linelog->path);
    }

    linelog->path = g_strdup (path);
}

void
linelog_disklog (Linelog *linelog, gboolean setting, GError **error)
{
    GError *err = NULL;

    if (setting)
    {
        if (linelog->logfile != NULL)
        {
            g_set_error (error, DM_DOWNMAND_ERROR, 1,
                "The linelog is already assigned to a file");
            return;
        }

        linelog->logfile = g_io_channel_new_file (linelog->path, "w", &err);

        if (err != NULL)
        {
            g_propagate_error (error, err);
            g_error_free (err);
        }
    }
    else
    {
        GIOStatus status;

        g_io_channel_flush (linelog->logfile, &err);

        if (err != NULL)
        {
            g_propagate_error (error, err);
            g_error_free (err);
            return;
        }

        status = g_io_channel_shutdown (linelog->logfile, TRUE, &err);

        if (err != NULL)
        {
            g_propagate_error (error, err);
            g_error_free (err);
        }

        linelog->logfile = 0;
    }
}
