#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include <glib.h>

#include <utillib.h>

#include "downmand-marshal.h"

#include "main.h"
#include "project.h"


GType           project_get_type            ();
static void     project_class_init          (ProjectClass *klass);
static void     project_init                (Project *project);

void        project_statistics_one      (gpointer key, gpointer value,
                                         gpointer user_data);
void        project_transfer_exit       (Transfer *transfer, gpointer data);
gboolean    project_transfer_real_exit  (GIOChannel *source,
                                         GIOCondition condition,
                                         gpointer data);

void    project_cleanup                     (void *arg);
void    project_session_batch_to_xml        (xmlNodePtr xmlnode, GList *batch,
                                             gchar *name);
void    project_session_transfer_to_xml     (xmlNodePtr xmlnode,
                                             Transfer *transfer);

gint    project_session_load                (Project *project);
void    project_session_xml_to_batch        (Project *project,
                                             xmlNodePtr xmlnode);
void    project_session_xml_to_transfer     (Project *project,
                                             xmlNodePtr xmlnode);

Transfer *  project_transfer_new        (Project *project, gchar *url);
void    project_transfer_connect_signals    (Project *project,
                                             Transfer *transfer);
void    project_transfer_update             (Transfer *transfer, gint setting,
                                             gpointer data);

void    project_transfer_links          (Project *project, Transfer *transfer);


enum {
    PROJECT_PATH,
    PROJECT_AUTOTRANSFER,
    PROJECT_MAXTRANSFERS,
    TRANSFER_NEW,
    TRANSFER_STOP,
    TRANSFER_DONE,
    TRANSFER_DELETE,
    TRANSFER_UPDATE,
    LAST_SIGNAL
};

static gint project_signals[LAST_SIGNAL] = { 0 };

static gpointer parent_class = NULL;


GType project_get_type ()
{
    static GType project_type = 0;

    if (!project_type)
    {
        static const GTypeInfo project_info =
        {
            sizeof (ProjectClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) project_class_init,
            NULL,           /* class_finalize */
            NULL,           /* class_data */
            sizeof (Project),
            0,              /* n_preallocs */
            (GInstanceInitFunc) project_init
        };

        project_type = g_type_register_static (G_TYPE_OBJECT,
                                               "Project",
                                               &project_info,
                                               0);
    }

    return project_type;
}

static void project_class_init (ProjectClass *klass)
{
    GObjectClass *object_class;

    object_class = (GObjectClass*) klass;

    parent_class = g_type_class_peek_parent (klass);

    project_signals[PROJECT_AUTOTRANSFER] = g_signal_new (
            "project_autotransfer_changed",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, project_autotransfer_changed),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    project_signals[PROJECT_MAXTRANSFERS] = g_signal_new (
            "project_maxtransfers_changed",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, project_maxtransfers_changed),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    project_signals[TRANSFER_NEW] = g_signal_new ("transfer_new",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, transfer_new),
            NULL, NULL,
            downmand_marshal_VOID__OBJECT,
            G_TYPE_NONE,
            1, transfer_get_type());

    project_signals[TRANSFER_STOP] = g_signal_new ("transfer_stop",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, transfer_stop),
            NULL, NULL,
            downmand_marshal_VOID__OBJECT,
            G_TYPE_NONE,
            1, transfer_get_type());

    project_signals[TRANSFER_DONE] = g_signal_new ("transfer_done",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, transfer_done),
            NULL, NULL,
            downmand_marshal_VOID__OBJECT,
            G_TYPE_NONE,
            1, transfer_get_type());

    project_signals[TRANSFER_DELETE] = g_signal_new ("transfer_delete",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, transfer_delete),
            NULL, NULL,
            downmand_marshal_VOID__OBJECT,
            G_TYPE_NONE,
            1, transfer_get_type());

    project_signals [TRANSFER_UPDATE] =
            g_signal_new ("transfer_update",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ProjectClass, transfer_update),
            NULL, NULL,
            downmand_marshal_VOID__OBJECT_INT,
            G_TYPE_NONE,
            2, transfer_get_type(), G_TYPE_INT);

    klass->project_autotransfer_changed = NULL;
    klass->project_maxtransfers_changed = NULL;
    klass->transfer_new = NULL;
    klass->transfer_stop = NULL;
    klass->transfer_done = NULL;
    klass->transfer_delete = NULL;
    klass->transfer_update = NULL;
}

static void
project_init (Project *project)
{
    project->hash = g_hash_table_new (NULL, NULL);
    project->hash_active = g_hash_table_new (NULL, NULL);
    project->batch = NULL;
    project->batch_active = NULL;
    project->transfer_id = 1;

    /* Counters */
    project->queued_transfers = 0;

    project->servers = servers_new ();
}

Project *
project_new (guint project_id, GError **error)
{
    Project *project = PROJECT (g_object_new (project_get_type (), NULL));

    project->project_id = project_id;

    /* default configuration */
    project->max_transfers = 2;
    project->auto_transfer = 1;

    return project;
}

void
project_scheduler (Project *project)
{
    GHashTable *hash_active = project->hash_active;
    GList *element;
    GError *err = NULL;
    GString *buffer = g_string_new (NULL);

//FIXME: Create a more abstracted scheduler
    // prepare for the posible iteration
    element = g_list_first (project->batch);

    while ((project->auto_transfer)
        && (g_hash_table_size (hash_active) < project->max_transfers)
        && (project->queued_transfers > 0)
        && (element != NULL))
    {
        Transfer *transfer;

        /* search for a queued transfer */
        while((element != 0))
        {
            transfer = TRANSFER (element->data);
            if (transfer_status_get (transfer) == STATUS_QUEUED)
            {
                /* check server limits */
//FIXME: check for source and target servers
//FIXME: check only if is_remote
                g_string_printf (buffer, "%s://%s",
                         gnome_vfs_uri_get_scheme (transfer->source_uri),
                         gnome_vfs_uri_get_host_name (transfer->source_uri));
                if (servers_connection_limit_reached (project->servers,
                                                      buffer->str))
                {
                    element = g_list_next (element);
                    continue;
                }
                else
                    break;
            }
            else
                element = g_list_next (element);
        }

        if (element != 0)
        {
            project->batch_active = g_list_append (project->batch_active,
                                                   transfer);
            g_hash_table_insert (hash_active,
                                 GINT_TO_POINTER (transfer_id_get (transfer)),
                                 transfer);
            transfer_queue_set (transfer, TRANSFER_QUEUE_ACTIVE);
            transfer_status_set (transfer, STATUS_INIT);
            project->queued_transfers--;

            // Start a new thread to transfer it
            transfer_start (transfer, &err);

            if (err == NULL)
            {
                servers_connection_start (project->servers, buffer->str);
            }
            else
            {
                transfer_queue_set (transfer, TRANSFER_QUEUE_STOPPED);
                transfer_status_set (transfer, STATUS_STOPPED);
                project->batch_active = g_list_remove (project->batch_active,
                                                       transfer);
                g_hash_table_remove (hash_active,
                                     GINT_TO_POINTER (transfer_id_get (transfer)));

g_print ("%s\n", err->message);
                g_error_free (err);
                err = NULL;
                g_signal_emit (G_OBJECT (project),
                               project_signals[TRANSFER_STOP],
                               0,
                               transfer);
            }

            // and move to the next element
            element = g_list_next (element);
        }
    }

    /* Harvest statistics */
    g_hash_table_foreach (hash_active, project_statistics_one, project);

    /* Do speed scheduling */
    if ((project->speedlimit) && (g_hash_table_size (hash_active)))
    {
        GList *transfers_list = NULL;
        GList *glist;
        Transfer *transfer;
        Transfer *slowest_transfer;
        guint active_transfers;
        guint bandwidth_left;
        guint bandwidth_piece;

        active_transfers = g_hash_table_size (hash_active);
        bandwidth_left = project->speedlimit;
        bandwidth_piece = project->speedlimit / active_transfers;

        /* copy the active transfers list */
        for (element = g_list_first (project->batch_active); element != 0;
             element = g_list_next (element))
        {
            transfers_list = g_list_prepend (transfers_list, element->data);
        }

        /* Remove slowest_transfer and keep recalculating bandwidth_piece */
        do {
            slowest_transfer = NULL;
            for (glist = g_list_first (transfers_list); glist != 0;
                 glist = g_list_next (glist))
            {
                transfer = (Transfer *) glist->data;
                if (transfer_speedlimit_get (transfer))
                {
                    if (slowest_transfer == NULL)
                        slowest_transfer = transfer;
                    else if (transfer_speedlimit_get (transfer)
                             < transfer_speedlimit_get (slowest_transfer))
                        slowest_transfer = transfer;
                }
            }

            if (slowest_transfer != NULL)
            {
                if (transfer_speedlimit_get (slowest_transfer)
                    <= bandwidth_piece)
                {
                    bandwidth_left -= transfer_speedlimit_get (
                                                            slowest_transfer);
                    transfers_list = g_list_remove (transfers_list,
                                                    slowest_transfer);
                    transfer_real_speedlimit_set (slowest_transfer,
                                    transfer_speedlimit_get (slowest_transfer));
                    /* recompute bandwidth_piece */
                    active_transfers--;
                    if (active_transfers)
                        bandwidth_piece = bandwidth_left / active_transfers;
                }
                else
                {
                    /* No more iterations needed, get out */
                    slowest_transfer = NULL;
                }
            }
        } while (slowest_transfer && active_transfers);

        /* look for already limited transfers */
        for (glist = g_list_first (transfers_list); glist != 0;
             glist = g_list_next (glist))
        {
            transfer = (Transfer *) glist->data;
            transfer_real_speedlimit_set (transfer, bandwidth_piece);
        }
    }
    else
    {
        Transfer *transfer;

        /* Set real_speedlimit on limited transfers and clear the rest */
        for (element = g_list_first (project->batch_active); element != 0;
             element = g_list_next (element))
        {
            transfer = element->data;
            if (transfer_speedlimit_get (transfer))
            {
                transfer_real_speedlimit_set (transfer,
                                            transfer_speedlimit_get (transfer));
            }
            else
                transfer_real_speedlimit_set (transfer, 0);
        }
    }

    g_string_free (buffer, TRUE);
}

void
project_statistics_one (gpointer key, gpointer value, gpointer user_data)
{
    transfer_statistics (value);
}

guint
project_id (Project *project)
{
    return project->project_id;
}

const char *project_name(Project *project)
{
	return project->name;
}

guint
project_transfer_id (Project *project)
{
    return project->transfer_id++;
}

gint
project_command_transfer_new (Project *project, gchar *command,
                              GError **error)
{
    Transfer *transfer;
    gchar **commands;
    gint i = 0;
    guint command_id = 0, value = 0;
    gboolean command_error = 0;

    transfer = project_transfer_new (project, NULL);

//FIXME: use GError here
	if (transfer == 0)
		return 1;

    transfer_log_add (transfer, "* Transfer created.\n");

    transfer_queue_set (transfer, TRANSFER_QUEUE_BATCH);
    transfer_status_set (transfer, STATUS_QUEUED);
    transfer_path_set (transfer, project->default_path);
    transfer_default_path_set (transfer, TRUE);

    commands = g_strsplit (command, "|", 0);

    for (command = commands[i]; command != NULL; command = commands[++i])
    {
        if (sscanf(command, "%u ", &command_id) == 0)
        {
            transfer_log_add (transfer, "* Bad command id: %s\n", command);
            continue;
        }

        switch (command_id)
        {
            /* Transfer commands */
            case DM_COMMAND_ITEM_URL:
//FIXME: invalid utf
//                transfer_log_add (transfer, "* Set url: %s\n", command + 4);
                if (!transfer_source_uri_set (transfer, command + 4))
                    command_error = 1;
                break;

            case DM_COMMAND_ITEM_PATH:
                transfer_log_add (transfer, "* Set PATH: %s\n", command + 4);
                transfer_path_set (transfer, command + 4);
                transfer_default_path_set (transfer, FALSE);
                break;

            case DM_COMMAND_ITEM_STOP:
                if (sscanf (command + 4, "%u", &value) > 0)
                {
                    if (value)
                    {
                        transfer_log_add (transfer, "* Set status: stopped.\n");
                        transfer_queue_set (transfer, TRANSFER_QUEUE_STOPPED);
                        transfer_status_set (transfer, STATUS_STOPPED);
                    }
                }
                break;

            case DM_COMMAND_ITEM_RECURSIVE:
                if (sscanf (command + 4, "%u", &value) > 0)
                {
                    transfer_log_add (transfer,
                                  "* Set FLAGS: Recursive on: %u.\n", value);
                    transfer_recursive_set (transfer, value);
                }
                break;

            case DM_COMMAND_ITEM_SPEED:
                if (sscanf (command + 4, "%u", &value) > 0)
                {
                    transfer_log_add (transfer,
                                  "* Set FLAGS: Speed limit on: %u.\n", value);
                    transfer_speedlimit_set (transfer, value);
                }
                break;

            default:
                transfer_log_add (transfer, "COMMAND NOT UNDERSTOOD: %s\n", command);
                break;
        }
    }

    g_strfreev(commands);

    if (command_error)
        return 1;

    transfer_log_add (transfer, "* Added to project list.\n");
	project->batch = g_list_append (project->batch, transfer);
    g_hash_table_insert (project->hash,
                         GINT_TO_POINTER (transfer_id_get (transfer)),
                         transfer);
    if (transfer_status_get (transfer) == STATUS_QUEUED)
        project->queued_transfers++;

    /* connect signals last so they don't get emitted */
    project_transfer_connect_signals (project, transfer);

    g_signal_emit (G_OBJECT (project), project_signals[TRANSFER_NEW], 0,
                   transfer);

	project_scheduler(project);

	return 0;
}

void
project_command_transfer_speed (Project *project, guint transfer_id,
                                guint setting,
                                GError **err)
{
    GHashTable *hash = project->hash;
    Transfer *transfer;

    transfer = g_hash_table_lookup (hash, GINT_TO_POINTER (transfer_id));
    if (transfer)
    {
        transfer_speedlimit_set (transfer, setting);
    }
}

/* This function is called when a transfer exits.
 */
void
project_transfer_exit (Transfer *transfer, gpointer data)
{
    Project *project = PROJECT (data);
	GHashTable *hash = project->hash;
	GHashTable *hash_active = project->hash_active;
    GError *err = NULL;
    GString *buffer = g_string_new (NULL);

    /* Update servers object */
    g_string_printf (buffer, "%s://%s",
                     gnome_vfs_uri_get_scheme (transfer->source_uri),
                     gnome_vfs_uri_get_host_name (transfer->source_uri));
    servers_connection_end (project->servers, buffer->str);

    switch (transfer_exit_cause(transfer))
    {
        case TRANSFER_EXIT_STOPPED:
            project->batch_active = g_list_remove (project->batch_active, transfer);
            g_hash_table_remove (hash_active,
                             GINT_TO_POINTER (transfer_id_get (transfer)));
            transfer_queue_set (transfer, TRANSFER_QUEUE_STOPPED);
            g_signal_emit (G_OBJECT (project),
                           project_signals[TRANSFER_STOP],
                           0,
                           transfer);
            break;

        case TRANSFER_EXIT_QUEUED:
            project->batch_active = g_list_remove (project->batch_active,
                                                   transfer);
            g_hash_table_remove (hash_active,
                             GINT_TO_POINTER (transfer_id_get (transfer)));
            transfer_queue_set (transfer, TRANSFER_QUEUE_BATCH);
            g_signal_emit (G_OBJECT (project),
                           project_signals[TRANSFER_STOP],
                           0,
                           transfer);
            break;

        case TRANSFER_EXIT_DELETE:
            project->batch_active = g_list_remove (project->batch_active,
                                                   transfer);
            g_hash_table_remove (hash_active,
                             GINT_TO_POINTER (transfer_id_get (transfer)));

            project->batch = g_list_remove (project->batch, transfer);
            g_hash_table_remove (hash,
                        GINT_TO_POINTER (transfer_id_get (transfer)));

            g_signal_emit (G_OBJECT (project),
                           project_signals[TRANSFER_DELETE],
                           0,
                           transfer);
            g_object_unref (G_OBJECT (transfer));
            break;

        case TRANSFER_EXIT_DONE:
            transfer_log_add (transfer, "* Done. Move to done.\n");
            project->batch_active = g_list_remove (project->batch_active,
                                                   transfer);
            g_hash_table_remove (hash_active,
                            GINT_TO_POINTER (transfer_id_get (transfer)));
            transfer_queue_set (transfer, TRANSFER_QUEUE_DONE);
            g_signal_emit (G_OBJECT (project),
                           project_signals[TRANSFER_DONE],
                           0,
                           transfer);

            if (transfer->recursive)
                project_transfer_links (project, transfer);
            break;

        case TRANSFER_EXIT_SHUTDOWN:
            transfer_log_add (transfer, "* Downmand shutdown.\n");
            break;

        default:
            transfer_log_add (transfer, "* No reason to exit. Trying again.\n");
//FIXME: check for number of retries
            transfer_start (transfer, &err);

            if (err == NULL)
            {
                servers_connection_start (project->servers, buffer->str);
            }
            else
            {
                transfer_queue_set (transfer, TRANSFER_QUEUE_STOPPED);
                transfer_status_set (transfer, STATUS_STOPPED);
                project->batch_active = g_list_remove (project->batch_active,
                                                       transfer);
                g_hash_table_remove (hash_active,
                            GINT_TO_POINTER (transfer_id_get (transfer)));

                g_error_free (err);
                g_signal_emit (G_OBJECT (project),
                               project_signals[TRANSFER_STOP],
                               0,
                               transfer);
            }

            break;
    }

    g_string_free (buffer, TRUE);
    return;
}

gint
project_command_transfer_start (Project *project, guint id, GError **error)
{
    GHashTable *hash = project->hash;
    GHashTable *hash_active = project->hash_active;
    Transfer *transfer;
    GError *err = NULL;
    GString *buffer = g_string_new (NULL);

    transfer = g_hash_table_lookup (hash, GINT_TO_POINTER (id));
    if (transfer)
    {
        if (transfer->source_uri)
        {
            switch (transfer_queue_get (transfer))
            {
                case TRANSFER_QUEUE_BATCH:
                    project->queued_transfers--;
                case TRANSFER_QUEUE_STOPPED:
                    project->batch_active = g_list_append (
                                                    project->batch_active,
                                                    transfer);
                    g_hash_table_insert (hash_active,
                                GINT_TO_POINTER (transfer_id_get (transfer)),
                                transfer);
                    transfer_queue_set (transfer, TRANSFER_QUEUE_ACTIVE);
                    transfer_start (transfer, &err);

                    if (err == NULL)
                    {
                        g_string_printf (buffer, "%s://%s",
                            gnome_vfs_uri_get_scheme (transfer->source_uri),
                            gnome_vfs_uri_get_host_name (transfer->source_uri));
                        servers_connection_start (project->servers, buffer->str);
                    }
                    else
                    {
                        transfer_queue_set (transfer, TRANSFER_QUEUE_STOPPED);
                        transfer_status_set (transfer, STATUS_STOPPED);
                        project->batch_active = g_list_remove (
                                                        project->batch_active,
                                                        transfer);
                        g_hash_table_remove (hash_active,
                                GINT_TO_POINTER (transfer_id_get (transfer)));

g_print ("%s\n", err->message);
                        g_error_free (err);
                    }
                    break;

            default:
                break;
            }
        }
    }

    return 0;
}

gint
project_command_transfer_stop (Project *project, guint id, GError **error)
{
    GHashTable *hash = project->hash;
    Transfer *transfer;

    transfer = g_hash_table_lookup (hash, GINT_TO_POINTER (id));
    if (transfer)
    {
        switch (transfer_queue_get (transfer))
        {
            case TRANSFER_QUEUE_BATCH:
                transfer_log_add (transfer, "* Stop. Move from batch to stopped.\n");
                if (transfer_status_get(transfer) == STATUS_QUEUED)
                    project->queued_transfers--;
                transfer_status_set (transfer, STATUS_STOPPED);
                transfer_queue_set (transfer, TRANSFER_QUEUE_STOPPED);
                g_signal_emit (G_OBJECT (project),
                               project_signals[TRANSFER_STOP],
                               0,
                               transfer);
                break;

            case TRANSFER_QUEUE_ACTIVE:
                transfer_log_add (transfer, "* Stopping the transfer.\n");
                transfer_stop (transfer, TRANSFER_STOP_NONE);
                break;

            default:
                transfer_log_add (transfer, "* Stop: can't be stopped.\n");
                break;
        }
    }

    return 0;
}

gint
project_command_transfer_queue (Project *project, guint id, GError **error)
{
    GHashTable *hash = project->hash;
    Transfer *transfer;

    transfer = g_hash_table_lookup (hash, GINT_TO_POINTER (id));

    if (transfer)
    {
        switch (transfer_queue_get (transfer))
        {
            case TRANSFER_QUEUE_STOPPED:
                transfer_log_add (transfer, "* Queue. Move from stopped to batch.\n");
                project->queued_transfers++;
                transfer_status_set (transfer, STATUS_QUEUED);
                transfer_queue_set (transfer, TRANSFER_QUEUE_BATCH);
                g_signal_emit (G_OBJECT (project),
                               project_signals[TRANSFER_STOP],
                               0,
                               transfer);
                break;

            case TRANSFER_QUEUE_ACTIVE:
                transfer_log_add (transfer, "* Queuing the transfer.\n");
                project->queued_transfers++;
                transfer_stop (transfer, TRANSFER_STOP_QUEUE);
                break;

            default:
                transfer_log_add (transfer, "* Queue: can't be queued.\n");
                break;
        }
    }

    return 0;
}

gint
project_command_transfer_delete (Project *project, guint id, GError **error)
{
    GHashTable *hash = project->hash;
    Transfer *transfer;

    transfer = g_hash_table_lookup (hash, GINT_TO_POINTER (id));

    if (transfer)
    {
        switch (transfer_queue_get (transfer))
        {
            case TRANSFER_QUEUE_BATCH:
                if (transfer_status_get (transfer) == STATUS_QUEUED)
                    project->queued_transfers--;
            case TRANSFER_QUEUE_STOPPED:
                transfer_log_add (transfer, "* Delete. Removing from project.\n");
                project->batch = g_list_remove (project->batch, transfer);
                g_hash_table_remove (hash, GINT_TO_POINTER (id));

                g_signal_emit (G_OBJECT (project),
                               project_signals[TRANSFER_DELETE],
                               0,
                               transfer);

                g_object_unref (G_OBJECT (transfer));
                break;

            case TRANSFER_QUEUE_ACTIVE:
                transfer_log_add (transfer, "* Deleting.\n");
                transfer_stop (transfer, TRANSFER_STOP_DELETE);
                break;

            case TRANSFER_QUEUE_DONE:
                transfer_log_add (transfer, "* Delete. Removing from project.\n");
                project->batch = g_list_remove (project->batch, transfer);
                g_hash_table_remove (hash, GINT_TO_POINTER (id));
                g_signal_emit (G_OBJECT (project),
                               project_signals[TRANSFER_DELETE],
                               0,
                               transfer);
                g_object_unref (G_OBJECT (transfer));
                break;

            default:
                transfer_log_add (transfer, "* Delete: invalid queue.\n");
                break;
        }
    }
    else
    {
//FIXME: log to project linelog
        g_print("Transfer id %i not found, so not deleted!\n", id);
    }

    return 0;
}

void
transfer_link (gpointer key, gpointer value, gpointer data)
{
    Project *project = PROJECT (data);
    Transfer *transfer;
    gchar *url = (gchar *) key;
    guint recursive;

    recursive = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (project),
                                                    "recursive"));

g_print("%s\n", url);
    transfer = project_transfer_new (project, url);
    transfer_queue_set (transfer, TRANSFER_QUEUE_BATCH);
    transfer_status_set (transfer, STATUS_QUEUED);
    transfer_path_set (transfer, project->default_path);
    transfer_default_path_set (transfer, TRUE);
    transfer_recursive_set (transfer, recursive);

	project->batch = g_list_append (project->batch, transfer);
    g_hash_table_insert (project->hash,
                         GINT_TO_POINTER (transfer_id_get (transfer)),
                         transfer);
    if (transfer_status_get (transfer) == STATUS_QUEUED)
        project->queued_transfers++;

    /* connect signals last so they don't get emitted */
    project_transfer_connect_signals (project, transfer);

    g_signal_emit (G_OBJECT (project), project_signals[TRANSFER_NEW], 0,
                   transfer);
}

void
project_transfer_links (Project *project, Transfer *transfer)
{
    GHashTable *hash;
//FIXME    GHashTable *hash_filtered;
    guint recursive = transfer->recursive;

    if (recursive-- == 0)
        return;

    hash = parse_html_file (transfer->target_uri, transfer->source_uri);

//FIXME: do this for all patterns
//    move_matching_strings (hash, hash_filtered, pattern);

    g_object_set_data (G_OBJECT (project), "recursive",
                       GINT_TO_POINTER (recursive));
    g_hash_table_foreach (hash, transfer_link, project);
}

gint
project_session_save (Project *project)
{
    GList *batch = project->batch;
    xmlDocPtr doc;
    GString *buffer = g_string_new (NULL);

    doc = xmlNewDoc ("1.0");
    doc->children = xmlNewDocNode (doc, NULL, "TransferManager", NULL);

    project_session_batch_to_xml (doc->children, batch, "batch");

//FIXME: save to a different file and test for integrity
    g_string_printf (buffer, "%s/batch", project->path);

    if (xmlSaveFile (buffer->str, doc) == -1)
    {
// FIXME: better error handling
        return 1;
    }

    xmlFreeDoc (doc);
    g_string_free (buffer, TRUE);

    return 0;
}

void
project_session_batch_to_xml (xmlNodePtr xmlnode, GList *batch, char *name)
{
	Transfer *transfer;
	GList *element;
	xmlNodePtr newxml;

	newxml = xmlNewChild(xmlnode, NULL, name, NULL);

	element = g_list_first (batch);
	while (element != 0)
	{
		transfer = element->data;

		project_session_transfer_to_xml (newxml, transfer);

		element = g_list_next (element);
	}
}

void
project_session_transfer_to_xml (xmlNodePtr xmlnode, Transfer *transfer)
{
    xmlNodePtr newxml;
    GString *buffer = g_string_new(NULL);
    gchar *uri;

    newxml = xmlNewChild (xmlnode, NULL, "Transfer", NULL);

    g_string_printf (buffer, "%u", transfer_queue_get (transfer));
    xmlSetProp (newxml, "queue", buffer->str);

    g_string_printf (buffer, "%u", transfer_status_get (transfer));
    xmlSetProp (newxml, "status", buffer->str);

    g_string_printf (buffer, "%u", transfer_recursive_get (transfer));
    xmlSetProp (newxml, "recursive", buffer->str);

    g_string_printf (buffer, "%u", transfer_speedlimit_get (transfer));
    xmlSetProp (newxml, "speedlimit", buffer->str);

    uri = transfer_source_uri_string (transfer);
    if (uri)
        xmlSetProp (newxml, "source_uri", uri);
    g_free (uri);

    uri = transfer_target_uri_string (transfer);
    if (uri)
        xmlSetProp (newxml, "target_uri", uri);
    g_free (uri);

    g_string_printf (buffer, "%Lu", transfer_bytes_total_get (transfer));
    xmlSetProp (newxml, "bytes_total", buffer->str);

    g_string_printf (buffer, "%Lu", transfer_bytes_done_get (transfer));
    xmlSetProp (newxml, "bytes_done", buffer->str);

    g_string_printf (buffer, "%i", transfer_target_created_get (transfer));
    xmlSetProp (newxml, "target_created", buffer->str);

    g_string_free (buffer, TRUE);
}

int project_session_load(Project *project)
{
	xmlDocPtr doc;
	xmlNodePtr node;
	GString *buffer = g_string_new(NULL);

	g_string_printf(buffer, "%s/batch", project->path);
	doc = xmlParseFile(buffer->str);

	if (!doc)
		return 1;

	if (!doc->children || !doc->children->name
		|| strcasecmp(doc->children->name, "TransferManager") != 0)
	{
		xmlFreeDoc(doc);
		return 1;
	}

	for (node = doc->children->children; node != NULL; node = node->next)
	{
		project_session_xml_to_batch(project, node);
	}

	xmlFreeDoc(doc);
	g_string_free(buffer, TRUE);

	return 0;
}

void
project_session_xml_to_batch (Project *project, xmlNodePtr xmlnode)
{
    xmlNodePtr node;

    if (!xmlnode->name)
        return;

    if (strcasecmp (xmlnode->name, "batch") != 0)
    {
        g_warning ("Unknow queue\n");
        return;
    }

    for (node = xmlnode->children; node != NULL; node = node->next)
    {
        project_session_xml_to_transfer (project, node);
    }
}

void
project_session_xml_to_transfer (Project *project, xmlNodePtr xmlnode)
{
	Transfer *transfer;
	char *uri;
	TransferQueue queue;
	enum dm_transfer_status status;
	unsigned long int bytes_total, bytes_done;
    guint recursive = 0, speedlimit = 0;
    xmlChar *prop;
    GError *err = NULL;
    GString *buffer = g_string_new (NULL);
    gboolean target_created;

	if (!xmlnode->name || strcasecmp(xmlnode->name, "Transfer") != 0)
		return;

	uri = xmlGetProp (xmlnode, "source_uri");
	if (uri == 0)
		return;
	
	queue = atol(xmlGetProp(xmlnode, "queue"));
	status = atol(xmlGetProp(xmlnode, "status"));

	prop = xmlGetProp (xmlnode, "recursive");
    if (prop != NULL)
    {
        recursive = atol (prop);
        xmlFree (prop);
    }

    prop = xmlGetProp(xmlnode, "speedlimit");
    if (prop != NULL)
    {
        speedlimit = atol (prop);
        xmlFree (prop);
    }

	transfer = project_transfer_new (project, uri);
	if (transfer == 0)
		return;

    prop = xmlGetProp(xmlnode, "bytes_total");
    if (prop != NULL)
    {
        bytes_total = atol(prop);
        transfer_bytes_total_set (transfer, bytes_total);
        xmlFree(prop);
    }

//FIXME: if it is a normal transfer, check that it equals the size of the file
    prop = xmlGetProp(xmlnode, "bytes_done");
    if (prop != NULL)
    {
        bytes_done = atol(prop);
        transfer_bytes_done_set (transfer, bytes_done);
        xmlFree(prop);
    }

    prop = xmlGetProp(xmlnode, "target_created");
    if (prop != NULL)
    {
        target_created = FALSE;
        target_created = atol (prop);
        transfer_target_created_set (transfer, target_created);
        xmlFree(prop);
    }

    transfer_queue_set (transfer, queue);
    transfer_status_set (transfer, status);
    transfer_path_set (transfer, project->default_path);
    transfer_default_path_set (transfer, TRUE);
    transfer_recursive_set (transfer, recursive);
	transfer_speedlimit_set (transfer, speedlimit);
	project->batch = g_list_append (project->batch, transfer);
    g_hash_table_insert (project_hash (project),
                         GINT_TO_POINTER (transfer_id_get (transfer)),
                         transfer);

    /* connect signals last so they don't get emitted */
    project_transfer_connect_signals (project, transfer);

    switch (queue)
    {
        case TRANSFER_QUEUE_ACTIVE:
            project->batch_active = g_list_append (project->batch_active,
                                                   transfer);
            g_hash_table_insert (project->hash_active,
                                 GINT_TO_POINTER (transfer_id_get (transfer)),
                                 transfer);
            transfer_start (transfer, &err);

            if (err == NULL)
            {
                g_string_printf (buffer, "%s://%s",
                         gnome_vfs_uri_get_scheme (transfer->source_uri),
                         gnome_vfs_uri_get_host_name (transfer->source_uri));
                servers_connection_start (project->servers, buffer->str);
            }
            else
            {
                transfer_queue_set (transfer, TRANSFER_QUEUE_STOPPED);
                transfer_status_set (transfer, STATUS_STOPPED);
                project->batch_active = g_list_remove (project->batch_active,
                                                       transfer);
                g_hash_table_remove (project->hash_active,
                                GINT_TO_POINTER (transfer_id_get (transfer)));

                g_error_free (err);
            }
            break;

        case TRANSFER_QUEUE_BATCH:
            transfer_status_set (transfer, STATUS_QUEUED);
            project->queued_transfers++;
            break;

        case TRANSFER_QUEUE_STOPPED:
            transfer_status_set (transfer, STATUS_STOPPED);
            break;

        default:
            break;
    }
}

Transfer *project_transfer_new(Project *project, gchar *text_uri)
{
    Transfer *transfer = transfer_new ();

    if (text_uri)
    {
        if (!transfer_source_uri_set (transfer, text_uri))
            return NULL;
    }

    transfer_id_set (transfer, project_transfer_id(project));
    transfer_logdir_set (transfer, project->path);

    return transfer;
}

void
project_transfer_connect_signals (Project *project, Transfer *transfer)
{
    if (transfer != NULL)
    {
        g_signal_connect (transfer, "transfer_update",
                          G_CALLBACK (project_transfer_update),
                          project);

        g_signal_connect (transfer, "transfer_exit",
                          G_CALLBACK (project_transfer_exit),
                          project);
    }
}

/* This function is called when a transfer changes its configuration.
 */
void
project_transfer_update (Transfer *transfer, gint setting, gpointer data)
{
    Project *project = PROJECT (data);
    GString *buffer = g_string_new (NULL);

    g_signal_emit (G_OBJECT (project), project_signals [TRANSFER_UPDATE], 0,
                   transfer,
                   setting);

    g_string_free (buffer, TRUE);
}

int project_preferences_save(Project *project)
{
	xmlDocPtr doc;
	xmlNodePtr newxml;
	GString *buffer = g_string_new(NULL);

//FIXME: modified flag --> do not write the config file if there are no modifications
	doc = xmlNewDoc("1.0");
	doc->children = xmlNewDocNode(doc, NULL, "TransferManagerPrefs", NULL);

	newxml = xmlNewChild(doc->children, NULL, "Settings", NULL);
    if (project->default_path)
    {
        xmlSetProp(newxml, "DefaultDir", project->default_path);
    }
    else
    {
        g_string_printf(buffer, "%s/transfers/%s/", g_get_home_dir(),
                        project->name);
        xmlSetProp(newxml, "DefaultDir", buffer->str);
    }

	g_string_printf(buffer, "%i", project->max_transfers);
	xmlSetProp(newxml, "MaxTransfers", buffer->str);

	g_string_printf(buffer, "%i", project->auto_transfer);
	xmlSetProp(newxml, "AutoTransfer", buffer->str);

//FIXME: save to a different file and test for integrity
	g_string_printf(buffer, "%s/prefs", project->path);

	if (xmlSaveFile(buffer->str, doc) == -1)
	{
// FIXME: better error handling
		return 1;
	}
	xmlFreeDoc(doc);

	return 0;
}

int project_preferences_load(Project *project)
{
	xmlDocPtr doc;
	xmlNodePtr node;
	GString *buffer = g_string_new(NULL);
	xmlChar *prefs;

	g_string_printf(buffer, "%s/prefs", project->path);
	doc = xmlParseFile(buffer->str);

	if (!doc)
		return 1;

	if (!doc->children || !doc->children->name
		|| strcasecmp(doc->children->name, "TransferManagerPrefs") != 0)
	{
		xmlFreeDoc(doc);
		return 1;
	}

	node = doc->children->children;

	if (!node->name || strcasecmp(node->name, "Settings") != 0)
		return 1;

	prefs = xmlGetProp(node, "DefaultDir");
    if (prefs)
    {
        project->default_path = prefs;
    }
    else
    {
        g_string_printf(buffer, "%s/transfers/%s/", g_get_home_dir(),
                        project->name);
        project->default_path = g_strdup(buffer->str);
    }

	prefs = xmlGetProp(node, "MaxTransfers");
	if (prefs)
    {
		project->max_transfers = (unsigned int) atoi(prefs);
        xmlFree(prefs);
    }
	else
    {
		project->max_transfers = 1;
    }

	prefs = xmlGetProp(node, "AutoTransfer");
	if (prefs)
    {
		project->auto_transfer = (unsigned int) atoi(prefs);
        xmlFree(prefs);
    }
	else
    {
		project->auto_transfer = 1;
    }

	xmlFreeDoc(doc);
	g_string_free(buffer, TRUE);

	return 0;
}

void
project_auto_set(Project *project, unsigned int setting, GError **error)
{
	project->auto_transfer = setting;
    g_signal_emit (G_OBJECT (project),
                   project_signals[PROJECT_AUTOTRANSFER],
                   0);
}

void
project_maxtransfers_set (Project *project, guint maxtransfers, GError **error)
{
	project->max_transfers = maxtransfers;
    g_signal_emit (G_OBJECT (project),
                   project_signals[PROJECT_MAXTRANSFERS],
                   0);
}

void
project_name_set (Project *project, const gchar *name)
{
    if (project->name)
        g_free (project->name);

    project->name = g_strdup (name);
}

const gchar*
project_name_get (Project *project)
{
    return project->name;
}

void
project_path_set (Project *project, const gchar *path)
{
    if (project->path)
        g_free (project->path);

    project->path = g_strdup (path);
}

const gchar*
project_path_get (Project *project)
{
    return project->path;
}

void
project_default_path_set (Project *project, const gchar *default_path)
{
    if (project->default_path)
        g_free (project->default_path);

    project->default_path = g_strdup (default_path);
}

const gchar*
project_default_path_get (Project *project)
{
    return project->default_path;
}

void
project_speedlimit_set (Project *project, guint speedlimit)
{
    project->speedlimit = speedlimit;
}

GHashTable *
project_hash (Project *project)
{
    return project->hash;
}

GList *
project_batch (Project *project)
{
    return project->batch;
}

GList *
project_batch_active (Project *project)
{
    return project->batch_active;
}

guint
project_size (Project *project)
{
    return g_hash_table_size (project->hash);
}

unsigned int project_queued_transfers(Project *project)
{
	return project->queued_transfers;
}

int project_auto(Project *project)
{
	return project->auto_transfer;
}

guint project_maxtransfers_get (Project *project)
{
	return project->max_transfers;
}
